/* Retrieve installed memory size from BIOS information */
/* http://dmtf.org/sites/default/files/standards/documents/DSP0134_2.8.0.pdf */
/* S.Chapeland / june 2007 */
/* Small modifications: E.Denes / april 2013 */   

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>

#define BYTE   unsigned char      // __u8
#define WORD   unsigned short     // __u16
#define DWORD  unsigned int       // __u32
#define QWORD  unsigned long long // __u64;


typedef struct {
  BYTE anchor[4];
  BYTE checksum;
  BYTE length;
  BYTE v_maj;
  BYTE v_min;
  WORD max_size;
  BYTE revision;
  BYTE formatted[5];
  BYTE i_anchor[5];
  BYTE i_checksum;
  WORD t_length;
  DWORD t_address;
  WORD n_tables;
  BYTE bcd_revision;
} smbios_entrypoint;

typedef struct {
  BYTE type;
  BYTE length;
  WORD handle;
} smbios_table_header;

typedef struct {
  BYTE type;
  BYTE length;
  WORD handle;
  BYTE vendor;
  BYTE bios_version;
  WORD bios_start_addr;
  BYTE bios_releaase_date;
  BYTE bios_rom_size;
  QWORD bios_char; 
  WORD bios_char_ext;
  BYTE sb_v_maj;
  BYTE sb_v_min;
  BYTE contr_maj;
  BYTE contr_min;
} smbios_table_mapbiosinfo;

typedef struct {
  BYTE type;
  BYTE length;
  WORD handle;
  DWORD address_start;
  DWORD address_end;
  WORD ma_handle;
  BYTE part_width;
} smbios_table_mapmem;

/*
typedef struct {
  BYTE type;
  BYTE length;
  WORD handle;
  DWORD address_start;
  DWORD address_end;
  WORD md_handle;
  WORD ma_handle;
  BYTE part_rowpos;
  BYTE part_interpos;
  BYTE part_interddepth;
} smbios_table_mapdev;
*/

#define DEVMEM_PATH "/dev/mem"


/* read physical memory at given address / length */
/* returns pointer to allocated copy (to be freed by user), or NULL on error */
void *read_mem(size_t address, size_t size) 
{
	int fd;
  void *buffer, *memmap;
	size_t offset;
	
	if((fd=open(DEVMEM_PATH, O_RDONLY))==-1) return NULL;
	if((buffer=malloc(size))==NULL)	return NULL;
	offset=address%getpagesize();
 	memmap=mmap(0, offset+size, PROT_READ, MAP_SHARED, fd, address-offset);
	printf("memmap = 0x%08lx, size: %ld => %p\n", address, size, memmap);
	if (memmap==(void *)-1) return NULL;
 	memcpy(buffer, memmap+offset, size);
	munmap(memmap, offset+size);
	close(fd);
	return buffer;
}

int main(int argc, char **argv)
{
  BYTE *mem=NULL;
  DWORD mem_addr,mem_size;
  smbios_entrypoint *sm=NULL;
  smbios_table_header *smth=NULL;
  void *map1=NULL,*map2=NULL;
  int i;
  
  // Map memory to find SMBIOS entry point
  mem_addr=0x000F0000;
  mem_size=0x000FFFFF-0x000F0000;
  mem=read_mem(mem_addr,mem_size);
  if (mem==NULL) 
  { 
    perror("read_mem"); 
    printf("Can not read memory\n");
    return 1;  }
  map1=mem;
  
  // Look for entry string
  for (i=0;i<mem_size;i+=16) {
    if (!strncmp(&mem[i],"_SM_",4)) {
      sm=(smbios_entrypoint *)&mem[i];
      break;
    }
  }
  if (sm==NULL) { printf("SMBIOS table not found\n"); return 1;}

  // Check table
  if (strncmp(sm->i_anchor,"_DMI_",5))
  {
    printf("SMBIOS table check failed, no _DMI_ found\n");
    return -1;
  }

  // checksumm
  int j,ck=0;
  for (j=0; j<sizeof(*sm); j++) 
    ck += ((BYTE *)(sm))[i]; 
  if (ck & 0xFF)
  {
    printf("SMBIOS table checksumm failed\n"); 
    return -1;
  } 
  printf("SMBIOS %d.%d - %d DMI tables\n",sm->v_maj,sm->v_min,sm->n_tables);

  // Map memory to read SMBIOS tables
  mem_addr=sm->t_address;
  mem_size=sm->t_length;
  mem=read_mem(mem_addr,mem_size);
  if (mem==NULL) { printf("can not read memory\n"); return 1;  }
  map2=mem;

  // Look for memory information
  unsigned long long memory_bytes=0;
  for (i=0;i<mem_size;) 
  {
    smth=(smbios_table_header *)&mem[i];
    i+=smth->length;
//    printf("type %d (%d bytes)\n",(int)smth->type,(int)smth->length);

    if (smth->type == 0) 
    {
      printf("type %d (%d bytes)\n",(int)smth->type,(int)smth->length);
      smbios_table_mapbiosinfo *smb;
      smb=(smbios_table_mapbiosinfo *)smth;
      printf("BIOS verion %d.%d\n", smb->sb_v_maj, smb->sb_v_min);
    }

    if (smth->type == 19) 
    {
      printf("type %d (%d bytes)\n",(int)smth->type,(int)smth->length);
      smbios_table_mapmem *smm;
      smm=(smbios_table_mapmem *)smth;
      DWORD msize;
      msize=smm->address_end - smm->address_start + 1;
      
      printf("\tStarting Address: 0x%08X%03X\n",
      smm->address_start>>2, (smm->address_start & 0x3)<<10);
      printf("\tEnding Address:   0x%08X%03X\n",
      smm->address_end>>2, ((smm->address_end & 0x3)<<10)+0x3FF);
        
      if ((msize & 0x000FFFFF)==0) 
        memory_bytes = (msize>>20) * ((unsigned long long)0x40000000); // GB
      else if((msize & 0x000003FF)==0)  
        memory_bytes = (msize>>10) * ((unsigned long long)0x100000);   // MB
      else 
        memory_bytes = (msize>>10) * ((unsigned long long)0x400);      // kB
//	break;
    } // type 19 

    // find table end  
    for (;i<mem_size-1;i++)
    {
      if (mem[i]==0) 
      {
        i++;
	if (mem[i]==0) 
        {
	  i++;
	  break;
	}   
      }
    } // table end
  } // table loop

  free(map1);
  free(map2);

  // return number of bytes on stdout (zero if no info found)
  printf("%llu\n",memory_bytes);

  return 0;  
}

