/*
**********************************************************************
** 
**  physmemDump.c
**
**  utility to dump the physmem memory of a selected physmem device
**  last modification 16-May-2013
**  by Ervin Denes
**
**********************************************************************
*/

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "physmem_lib.h"

/* utitlity functions */

void usage(void)
{
  printf("USAGE: physmemDump {-o <byte offset> | -O <word offset> | "
                            "{-p|-P} <phys address>}\n");
  printf("                   {-l <length in bytes> | -L <length in words>}\n");
  printf("                   [{-m|-M} <minor device number | 0] [{-d|-D} <word width> | 4]\n");
  printf("       byte address/offset is rounded to lower,\n");
  printf("       byte length to higher word boundary.\n");
  printf("       values staring with 0x are interpreted as hex values.\n");
  exit(EXIT_FAILURE);
}

unsigned long getarg(char *opt)
{
  return (strtoul(opt, (char**)NULL, 0));
}

int main (int argc, char *argv[])
{
  int i, j, c, ret;
  physmemHandler_t physDevice;
  unsigned long phys_size;
  unsigned long phys_addr_dump; 
  unsigned long phys_addr; 
  unsigned long *ptr_dump;
  unsigned long offset, aux_offset, length;
  unsigned long word_offset, word_length;
  int minor, data_width, address_width;
  int arch;

#define FULL_SIZE 0
 
  offset = -1;
  length = -1;
  phys_addr = -1; 
  minor  = 0;
  data_width = 4;
  arch = (sizeof(void *) > 4) ? 64 : 32;
  address_width = (arch == 64) ? 8: 4;

  if(argc > 1) {
    while((c = getopt(argc, argv, "O:o:P:p:L:l:M:m:D:d:")) != EOF)
      switch(c)
      {
      case 'O': offset = getarg(optarg) * address_width;   break;
      case 'o': offset = getarg(optarg);                   break;
      case 'P':
      case 'p': phys_addr = getarg(optarg);                break;
      case 'L': length = getarg(optarg) * address_width;   break;
      case 'l': length = getarg(optarg);                   break;
      case 'M': 
      case 'm': minor  = getarg(optarg);                   break;
      case 'D':
      case 'd': data_width = getarg(optarg);               break;
      default : usage();                                   break;
      }
  }

  if (((offset == -1) && (phys_addr == -1))|| (length == -1))
    usage();

  if ((data_width != 4) && (data_width != 8))
  { 
    printf("Data width should be 4 or 8 (bytes)\n");
    usage();
  }

  if ((arch == 32) && (data_width != 4))
  {
    printf("For 32-bit architecture only 4 byte data_width can be used\n");
    data_width = 4;
  }

  /* open the device for physmem memory */ 
  ret = physmemOpen(&physDevice, minor, FULL_SIZE, 0);

  if (ret)
    exit (EXIT_FAILURE);

  phys_size = FULL_SIZE == 0 ? physDevice.phys_size_physmem : 
                               physDevice.full_size_physmem;

  /* check the parameters */
  if (offset != -1)  
  {
    if (offset >= phys_size)
    {
      printf("The physical size of physmem is %ld bytes, offset (%ld) is too big\n", 
                 phys_size, offset);
      physmemClose(physDevice);
      exit (EXIT_FAILURE);
    }
  }

  /* print zones */
  physmemPrintZones(physDevice);

  /* calculate the user offset */
  if (phys_addr != -1)
  {
    offset = physmemUserOffset(phys_addr, physDevice);

    if (offset == -1)
    {
      printf("\nPhysical address %#lx does not correspond to a user offset in physmem%d device\n", 
                                           phys_addr, minor);
      physmemClose(physDevice);
      exit (EXIT_FAILURE);
    }
    printf("\nPhysical address: %#lx, user offset: %#lx\n", phys_addr, offset);
  }

  word_offset = offset / address_width; // offset in long words
  offset = word_offset * address_width; // offset in bytes rounded to address_width

  word_length = length/address_width;
  if (word_length * address_width != length)
    word_length++;
  length = word_length * address_width;

  if ((offset + length) > phys_size)
  {
    length = phys_size - offset;
    word_length = length/address_width;

    printf("The physical size of physmem is %ld bytes, length is truncated to %ld bytes\n", 
               phys_size, length);
  }

  if (length)
  {
    /* calculate the start of user address */
    ptr_dump = (unsigned long *)physDevice.addr_user_physmem;
    ptr_dump += word_offset; // add the offset in long words

    /* calculate the start of physical address */
    aux_offset = offset;
    phys_addr_dump = physDevice.phys_addr_physmem + aux_offset;
    for (j = 0; j < physDevice.num_zones - 1; j++)
    { 
      if (phys_addr_dump >= physDevice.zones[j].end)
      {
        phys_addr_dump = physDevice.zones[j+1].start;
        aux_offset -= physDevice.zones[j].size;
        phys_addr_dump += aux_offset;
      }
    }

    /* dump */
    printf("\nDump of device %s (size: 0x%lx = %ld bytes (%ld MB))\n", 
	   physDevice.physmem_dev, 
           phys_size, phys_size, phys_size/1024/1024);
    if (arch == 32)
    {
      printf(
"--------------------------------------------------------------------------- \n");
      printf(
"        address                     offset                %d byte-word      \n", data_width);
      printf(
" physical:       user:       hex:            dec:      hex:            dec: \n");
      printf(
"--------------------------------------------------------------------------- \n");
    }
    else
    {
      if (data_width == 4)
      {
        printf(
"------------------------------------------------------------------------------------------------------------ \n");
        printf(
"                address                                       offset                        %d byte-word      \n", data_width);
        printf(
"     physical:               user:                 hex:                    dec:           hex:       dec: \n");
        printf(
"------------------------------------------------------------------------------------------------------------- \n");
      }
      else
      {
        printf(
"------------------------------------------------------------------------------------------------------------------------------ \n");
        printf(
"                address                                       offset                                 %d byte-word      \n", data_width);
        printf(
"     physical:               user:                   hex:                  dec:             hex:                       dec: \n");
        printf(
"------------------------------------------------------------------------------------------------------------------------------- \n");
      }
    }
  }

  for (i = 0; i < word_length; i++)
  {
    if (address_width == 4)
    {
       printf("0x%08x   0x%08lx   0x%08lx   %10ld   0x%08x   %10d \n",
             (unsigned int)phys_addr_dump,
             (datePointer)ptr_dump,
             offset, offset,
             (unsigned int)*ptr_dump, (unsigned int)*ptr_dump);
    }
    else
    { 
      if (data_width == 8)
      {
        printf("0x%016lx   0x%016lx   0x%016lx   %20ld   0x%016lx   %20ld \n",
               phys_addr_dump,
               (unsigned long)ptr_dump,
               offset, offset,
               *ptr_dump, *ptr_dump);
      }
      else
      {
        printf("0x%016lx   0x%016lx   0x%016lx   %20ld   0x%08x   %10d \n",
               phys_addr_dump,
               (unsigned long)ptr_dump,
               offset, offset,
               (int)*ptr_dump, (int)*ptr_dump);
        printf("0x%016lx   0x%016lx   0x%016lx   %20ld   0x%08x   %10d \n",
               phys_addr_dump + 4,
               (unsigned long)ptr_dump + 4,
               offset + 4, offset + 4,
               *(int *)((unsigned long)ptr_dump + 4), 
               *(int *)((unsigned long)ptr_dump + 4));
      }
    }
    ptr_dump++;
    phys_addr_dump += address_width;
    for (j = 0; j < physDevice.num_zones - 1; j++)
    { 
      if ((phys_addr_dump >= physDevice.zones[j].end) && 
          (phys_addr_dump < physDevice.zones[j+1].start))
      {
        phys_addr_dump = physDevice.zones[j+1].start;
        printf("\n                %ld bytes (%ld MB) 'hole' in physmem memory\n\n", 
            physDevice.zones[j+1].start - physDevice.zones[j].end,
            (physDevice.zones[j+1].start - physDevice.zones[j].end)/1024/1024);
        break;
      }
    }
    offset += address_width;
  }
  printf("\n");

  /* close the device for physmem */
  physmemClose(physDevice);

  exit(EXIT_SUCCESS);
}
