/*
**********************************************************************
** 
**  physmemFill.c
**
**  utility to fill the physmem memory under selected device
**  last modification 16-May-2013
**  by Ervin Denes
**
**********************************************************************
*/

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "physmem_lib.h"

#define BIT32 (unsigned long)0x0ffffffff

/* utitlity functions */

void usage(void)
{
  printf("USAGE: physmemFill {-o <byte offset> | -O <word offset> | "
                            "{-p|-P} <phys address>}\n");
  printf("                   {-l <length in bytes> | -L <length in words>}\n");
  printf("                   [{-s|-S} <start value> | 0] [{-i|-I} <increase value> | 1]\n");
  printf("                   [{-m|-M} <minor device number | 0] [{-d|-D} <word width> | 4]\n");
  printf("       byte offset is rounded to lower, byte length to higher word boundary.\n");
  printf("       values staring with 0x are interpreted as hex values.\n");
  exit(EXIT_FAILURE);
}

unsigned long getarg(char *opt)
{
  return (strtoul(opt, (char**)NULL, 0));
}

int main (int argc, char *argv[])
{
  int i, c, ret;
  physmemHandler_t physDevice;
  unsigned long phys_size;
  unsigned long phys_addr;
  unsigned long offset, length;
  unsigned long word_offset, word_length;
  unsigned long fill_value;
  unsigned long *ptr_fill;
  unsigned long start, incr;
  int minor, data_width, address_width;
  int arch;

  #define FULL_SIZE 0

  offset = -1;
  phys_addr = -1;
  length = -1;
  start = 0;
  incr = 1;
  minor = 0;
  data_width = 4;
  arch = (sizeof(void *) > 4) ? 64 : 32;
  address_width = (arch == 64) ? 8: 4;

  if(argc > 1) 
    while((c = getopt(argc, argv, "O:o:P:p:L:l:S:s:I:i:M:m:D:d:")) != EOF)
      switch(c)
      {
      case 'O': offset = getarg(optarg) * address_width; break;
      case 'o': offset = getarg(optarg);                 break;
      case 'P': 
      case 'p': phys_addr = getarg(optarg);              break;
      case 'L': length = getarg(optarg) * address_width; break;
      case 'l': length = getarg(optarg);                 break;
      case 'S':
      case 's': start  = getarg(optarg);                 break;
      case 'I':
      case 'i': incr   = getarg(optarg);                 break;
      case 'M': 
      case 'm': minor  = getarg(optarg);                 break;
      case 'D':
      case 'd': data_width = getarg(optarg);             break;

      default : usage();                                 break;
      }

  if (((offset == -1) && (phys_addr == -1))|| (length == -1))
    usage();

  if ((data_width != 4) && (data_width != 8))
  {
    printf("Data width should be 4 or 8 (bytes)\n");
    usage();
  }

  if ((arch == 32) && (data_width != 4))
  {
    printf("For 32-bit architecture only 4 byte data_width can be used\n");
    data_width = 4;
  }

  if (data_width == 4)
  {
    if (start > BIT32)
    {
      start &= BIT32;
      printf("Start fill value truncated to 32 bits\n");
    }
    if (incr > BIT32)
    {
      incr &= BIT32;
      printf("Increment value truncated to 32 bits\n");
    }
  }

  /* open the device for physmem memory */ 
  ret = physmemOpen(&physDevice, minor, FULL_SIZE, 0);

  if (ret)
    exit (EXIT_FAILURE);

  phys_size = FULL_SIZE == 0 ? physDevice.phys_size_physmem :
                               physDevice.full_size_physmem;

  /* check the parameters */
  if (offset != -1)
  {
    if (offset >= phys_size)
    {
      printf("The physical size of physmem is %ld bytes, offset (%lu) is too big\n", 
                 phys_size, offset);
      physmemClose(physDevice);
      exit (EXIT_FAILURE);
    }
  }

  /* calculate the user offset */
  if (phys_addr != -1)
  {
    offset = physmemUserOffset(phys_addr, physDevice);

    if (offset == -1)
    {
      printf("\nPhysical address %#lx does not correspond to a logical offset in physmem%d device\n",
                                           phys_addr, minor);
      physmemClose(physDevice);
      exit (EXIT_FAILURE);
    }
    printf("\nPhysical address: %#lx, logical offset: %#lx\n", phys_addr, offset);
  }

  word_offset = offset / address_width; // offset in long words
  offset = word_offset * address_width; // offset in bytes rounded to address_width

  word_length = length/address_width;
  if (word_length * address_width != length)
    word_length++;
  length = word_length * address_width;

  if ((offset + length) > phys_size)
  {
    length = phys_size - offset;
    word_length = length/address_width;

    printf("The physical size of physmem is %ld bytes, length is truncated to %ld bytes\n", 
               phys_size, length);
  }

  /* calculate the start of the user address */
  ptr_fill = (unsigned long *)physDevice.addr_user_physmem;
  ptr_fill += word_offset; // add the offset in longwords

  /* fill value */
  fill_value = start;

  for (i = 0; i < word_length; i++)
  {
    if (address_width == data_width)
    {
      *ptr_fill = fill_value;
      fill_value += incr;
    }
    else // address_width == 8 and data_width == 4
    {
      *ptr_fill = fill_value;
      fill_value += incr;
      *(int *)((unsigned long)ptr_fill + 4) = fill_value;
      fill_value += incr;
    }
    ptr_fill++;
  }
  printf("\n");
  if (length)
    printf("%ld (0x%08lx) words written to %s \n",
              (long)word_length, (long)word_length, physDevice.physmem_dev);
  else
   printf("Nothing to do\n");
  printf("\n");

  /* close the device for physmem */
  physmemClose (physDevice);

  exit(EXIT_SUCCESS);
}
