/*
**********************************************************************
** 
**  physmemFillWithAddress.c
**
**  utility to fill the physmem memory under device /dev/physmem
**  with its own physical addresses
**  last modification 16-May-2013
**  by Ervin Denes
**
**********************************************************************
*/

#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include "physmem_lib.h"

/*
** utitlity functions
*/

void usage(void)
{
  printf("physmemFillWithAddress: fill the full physmem memory \n");
  printf("                               with its own physical addresses.\n");
  printf("USAGE: physmemFillWithAddress [[-m|-M] <minor device number> | 0]\n");
  printf("                              [[-d|-D] <data width in bytes> | 4]\n");
  exit(EXIT_FAILURE);
}

/*
** entry point of test program
*/

int main (int argc, char *argv[])
{
  int i, j, c, ret;
  physmemHandler_t physDevice;
  unsigned long phys_addr; 
  unsigned long phys_size; 
  unsigned long fill_value;
  unsigned long *ptr_fill;
  unsigned long word_length;
  int minor, data_width, address_width;
  int arch;

  #define FULL_SIZE 0

  minor = 0;
  data_width = 4;
  arch = (sizeof(void *) > 4) ? 64 : 32;
  address_width = (arch == 64) ? 8: 4;

  if(argc > 1)
    while((c = getopt(argc, argv, "M:m:D:d:")) != EOF)
      switch(c)
      {
        case 'M':
        case 'm': minor  = atoi(optarg);      break;
        case 'D':
        case 'd': data_width  = atoi(optarg); break;
        default : usage();                    break;
      }

  if ((data_width != 4) && (data_width != 8))
  {
    printf("Data width should be 4 or 8 (bytes)\n");
    usage();
  }

  if ((arch == 32) && (data_width != 4))
  {
    printf("For 32-bit architecture only 4 byte data_width can be used\n");
    data_width = 4;
  }

  /* open the device for physmem memory */ 
  ret = physmemOpen(&physDevice, minor, FULL_SIZE, 0);

  if (ret)
    exit (EXIT_FAILURE);

  /* get size of the physmem */
  phys_size = FULL_SIZE == 0 ? physDevice.phys_size_physmem :
                               physDevice.full_size_physmem;

  /* user start address */
  ptr_fill = (unsigned long *)physDevice.addr_user_physmem;

  /* length */
  word_length = phys_size/address_width;

  /* fill value */
  phys_addr = physDevice.phys_addr_physmem;
  fill_value = phys_addr;

  for (i = 0; i < word_length; i++)
  {
    if (address_width == data_width)
    {
      *ptr_fill = fill_value;
    }
    else // address_width == 8 and data_width == 4
    {
      *ptr_fill = (fill_value & 0xffffffff);
      *(int *)((unsigned long)ptr_fill + 4) = (fill_value + 4) & 0xffffffff;
    }
    fill_value += address_width;
    ptr_fill++;

    phys_addr += address_width;
    for (j = 0; j < physDevice.num_zones - 1; j++)
    {
      if ((phys_addr >= physDevice.zones[j].end) && 
          (phys_addr < physDevice.zones[j+1].start))
      {
        phys_addr = physDevice.zones[j+1].start;
        break;
      }
    }
    fill_value = phys_addr;
  }
  printf("\n");
  printf("\n%ld words (%ld MB) of %s memory filled with own physical address\n",
               word_length, word_length*address_width/1024/1024, 
               physDevice.physmem_dev);
  printf("\n");

  /* close the device for physmem */
  physmemClose (physDevice);

  exit(EXIT_SUCCESS);
}
