/*
**********************************************************************
**
**  physmemTest.c
**
**  utility to test the physmem memory selected physmem device
**  last modification 25-Jan-2010
**  by Klaus Schossmaier, Marc Fauman and Ervin Denes
**
**********************************************************************
*/

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include "physmem.h"

#ifndef datePointer
#define datePointer long
#endif

#define PATTERN 2

void write_physmem (void*, long);
void read_physmem (void*, long);
void usage(void)
{
  printf("USAGE: physmemTest [{-m|-M} <physmem minor device number> | 1]\n");
  printf("                   [{-t|-T} <test method> | 3]\n");
  printf("                   [{-l|-L} <test length> | 1000]\n");
  printf("                   [{-o|-O} <test offset> | -1]\n");
  printf("   <test method> = 1: only GETSIZE test\n");
  printf("   <test method> = 2: only GETFULLSIZE test\n");
  printf("   <test method> = 3: both tests\n");
  printf("   <test offset>: offset in bytes where to test write and read\n");
  printf("                  -1 means at end of zone\n");  
  exit(EXIT_FAILURE);
}

/*----------------------------------------------------------*/
int main(int argc, char **argv)
{
  int c, minor, physmem_fd;
  void *addr_user_physmem;
  unsigned long phys_addr_physmem;
  unsigned long phys_size_physmem;
  unsigned long phys_full_size_physmem;
  unsigned int phys_num_zones;
  unsigned long tlength, toffset, tlength_korr;
  struct memZoneStruct *phys_zones;
  int j;
  char physmem_name[16];
  int testm;

  testm = 3;
  minor = 1;
  toffset = (unsigned long)-1;
  tlength = 1000;
  if(argc > 1)
 {
    while((c = getopt(argc, argv, "M:m:T:t:O:o:L:l:")) != EOF)
      switch(c)
      {
        case 'M':
        case 'm': minor = atoi(optarg);   break;
        case 'T':
        case 't': testm = atoi(optarg);   break;
        case 'O':
        case 'o': toffset = strtoul(optarg, (char**)NULL, 0);    break;
        case 'L':
        case 'l': tlength = atol(optarg); break;
        default : usage();                break;
      }
  }

  if ((minor < 0) || (minor >= PHYSMEM_MAXDEVICE))
  {
    printf("Physmem minor %d is not valid\n", minor);
    usage();
  }

  if ((testm < 1) || (testm > 3))
  {
    printf("Test method %d is not valid\n", testm);
    usage();
  }

  sprintf(physmem_name, "/dev/physmem%d", minor);
  if ((physmem_fd = open (physmem_name, O_RDWR)) == -1)
  {
    perror("open");
    printf("Cannot open device %s \n", physmem_name);
    exit (EXIT_FAILURE);
  }
  printf ("Opening physmem device %s\n", physmem_name);
  
  if (ioctl (physmem_fd, PHYSMEM_GETADDR, &phys_addr_physmem) < 0) {
    perror("ioctl");
    printf ("Cannot ioctl %s to get address\n", physmem_name);
    close (physmem_fd);
    exit (EXIT_FAILURE);
  }
  printf ("Physical address = %#lx\n", phys_addr_physmem);
  fflush (NULL);

  /* get and print zones */
  if (ioctl (physmem_fd, PHYSMEM_GETNUMMEMZONES, &phys_num_zones) < 0) {
    perror("ioctl");
    printf ("Cannot ioctl device %s to get number of useable blocks\n",
                                                             physmem_name);
    close (physmem_fd);
    exit (EXIT_FAILURE);
  }  
  printf (" + Number of memory zones : %d\n", phys_num_zones);
  fflush (NULL);

  phys_zones = (struct memZoneStruct *) malloc (phys_num_zones*sizeof (struct memZoneStruct));

  if (ioctl (physmem_fd, PHYSMEM_GETMEMZONES, phys_zones) < 0) {
    perror("ioctl");
    printf ("Cannot ioctl device %s to get memory zones \n", physmem_name);
    close (physmem_fd);
    exit (EXIT_FAILURE);
  }

  for (j=0; j < phys_num_zones; j++)
    printf (" + Mem zone %d : %#lx -> %#lx (size = %#lx)\n", 
            j+1, phys_zones[j].start, phys_zones[j].end, phys_zones[j].size);
  fflush (NULL);

  /* no hole test */

  if (testm & 0x1)  
  {
    if (ioctl (physmem_fd, PHYSMEM_GETSIZE, &phys_size_physmem) < 0){
      perror("ioctl");
      printf ("Cannot ioctl %s to get size\n", physmem_name);
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
    printf ("Physical usable size of device (no hole) : %#lx\n", phys_size_physmem);
    fflush (NULL);

    if (phys_size_physmem <= 0)
    {
      printf ("Can not test 0 size\n");
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
  
    printf ("mapp'ing\n");
    fflush (NULL);
    if ((addr_user_physmem = mmap (0, phys_size_physmem, PROT_READ|PROT_WRITE, MAP_SHARED, physmem_fd, 0)) == MAP_FAILED){
      perror ("mmap");
      printf ("Mmap failed\n");
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
    printf ("Mmap done to %p -> %p, trying to access physmem\n", 
            addr_user_physmem, addr_user_physmem+phys_size_physmem);
    fflush (NULL);
  
    write_physmem (addr_user_physmem, tlength);
    read_physmem (addr_user_physmem, tlength);

    if (phys_num_zones > 1)
    {  
      /* writing just after the "separation" in memory */  
      printf ("Writing just after the 'hole' in memory\n");
      write_physmem (addr_user_physmem+phys_zones[0].size, tlength);
      read_physmem (addr_user_physmem+phys_zones[0].size, tlength);
    }  
    else
    {
      printf ("Writing just before the memory end\n");
      write_physmem (addr_user_physmem+phys_zones[0].size-tlength, tlength);
      read_physmem (addr_user_physmem+phys_zones[0].size-tlength, tlength);
    }  
  }

  /* full size test */ 

  if (testm & 0x2)  
  {
    if (testm & 0x1)  
    {
      printf ("unmapp'ing\n");
      fflush (NULL);
      munmap (addr_user_physmem, phys_size_physmem);
    }

    if (ioctl (physmem_fd, PHYSMEM_GETFULLSIZE, &phys_full_size_physmem) < 0){
      perror ("ioctl");
      printf ("Cannot ioctl %s to get full size\n", physmem_name);
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }  
  
    printf ("physical full size of device (including hole) = %#lx\n", 
                                                     phys_full_size_physmem);
    fflush (NULL);

    if (phys_full_size_physmem <= 0)
    {
      printf ("Can not test 0 size\n");
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }
  
    if (!(testm & 0x1))
      printf ("mapp'ing\n");
    else
      printf ("re-mapp'ing\n");
    fflush (NULL);

    if ((addr_user_physmem = mmap (0, phys_full_size_physmem, PROT_READ|PROT_WRITE, MAP_SHARED, physmem_fd, 0)) == MAP_FAILED){
      perror ("mmap");
      printf ("Mmap failed\n");
      close (physmem_fd);
      exit (EXIT_FAILURE);
    }  
  
    printf ("Mmap done to %p -> %p, trying to access physmem\n", 
            addr_user_physmem, addr_user_physmem+phys_full_size_physmem);
    fflush (NULL);
    
    if (tlength > phys_zones[0].size)
    {
      tlength_korr = phys_zones[0].size;          
      printf ("Test length %ld greater than zone[0].size, truncated to %ld\n",
               tlength, tlength_korr);
    }
    else
      tlength_korr = tlength;

    if (!(testm & 0x1))
    {
      printf ("Writing to the start of memory\n");
      write_physmem (addr_user_physmem, tlength_korr);
    }  
    printf ("Reading from the start of memory\n");
    read_physmem (addr_user_physmem, tlength_korr);

    if (phys_num_zones > 1)
    {
      printf ("phys_zones[1].start = %#lx\n", phys_zones[1].start);
      printf ("hole user addr: from %p to %p\n",
               addr_user_physmem + phys_zones[0].size,
               addr_user_physmem + phys_zones[1].start - phys_zones[0].start);
      if (!(testm & 0x1))
      {  
        printf ("Writing just after the 'hole' in memory\n");
        write_physmem (addr_user_physmem+phys_zones[1].start-phys_zones[0].start,
                                                                      tlength);
      }
      printf ("Reading just after the 'hole' in memory\n");
      read_physmem (addr_user_physmem+phys_zones[1].start-phys_zones[0].start,
                                                                      tlength);
      if ((toffset != (unsigned long)-1) && 
          ((toffset - tlength) >= phys_zones[0].size) && 
          ((toffset + tlength) < phys_zones[1].start-phys_zones[0].start))
      {
        printf ("Offset 0x%lx points into the hole at %p. Not used\n", 
                               toffset, addr_user_physmem + toffset);
        toffset = (unsigned long)-1;
      }

      if ((toffset + tlength) > (phys_zones[1].end-phys_zones[0].start))
      {
        tlength_korr = phys_zones[1].end-phys_zones[0].start - toffset; 
        printf ("Test length %ld greater than zone[1].size, truncated to %ld\n",
                 tlength, tlength_korr);
      }  
      else
        tlength_korr = tlength;

      if (toffset == (unsigned long)-1)
      {
        printf ("Writing just before the phys_zone[1].end\n"); 
        write_physmem (addr_user_physmem+phys_zones[1].end-phys_zones[0].start-tlength_korr,
                                                                 tlength_korr);
        read_physmem (addr_user_physmem+phys_zones[1].end-phys_zones[0].start-tlength_korr,
                                                                 tlength_korr);
      }
      else
      {
        printf ("Writing at offset 0x%lx\n", toffset); 
        write_physmem (addr_user_physmem + toffset, tlength_korr);
        read_physmem  (addr_user_physmem + toffset, tlength_korr);
      }
      fflush (NULL);
close (physmem_fd);
exit (0);
      /* reading the "separation" in memory */  
      read_physmem (addr_user_physmem+phys_zones[0].size, tlength);
    }
    else
    {
      if (!(testm & 0x1))
      {
        printf ("Writing just before the memory end\n");
        write_physmem (addr_user_physmem+phys_zones[0].size-tlength, tlength);
      }
      printf ("Reading just before the memory end\n");
      read_physmem (addr_user_physmem+phys_zones[0].size-tlength, tlength);
    }
  }
  close (physmem_fd);
  exit (EXIT_SUCCESS);
}

/*-------------------------------------------------------------*/
/* function for writing to the physmem memory */

void write_physmem(void* addr_user_physmem, long length) 
{
  unsigned long *ptr_test;   
  int i;

  ptr_test = (unsigned long *)addr_user_physmem;

    /* Loop over all addresses (sizeof(long) = size of an address on 32 and 64-bit machines) */
  for (i=0; i<length/sizeof(long); i++){
    *(ptr_test+i) = PATTERN + i;
  }
  printf(" + test write: from %#lx, %ld bytes written \n", (datePointer)addr_user_physmem,length); 
}

/*-------------------------------------------------------------*/
/* function for reading from the physmem memory */

void read_physmem(void* addr_user_physmem, long length) 
{
  unsigned long *ptr_test;   
  int i;
  long first_int;
  int flag;

  ptr_test = (unsigned long *)addr_user_physmem;
  first_int = *ptr_test;
  if( first_int != PATTERN ) {
    printf("\n - pattern does not match: found %ld, expected %d \n\n", 
	   first_int, PATTERN);
  }
  else
  {
    printf(" + test read:  from %#lx, ", (datePointer)addr_user_physmem); 
    flag = 1; // initially ok

      /* Loop over all addresses (sizeof(long) = size of an address on 32 and 64-bit machines) */
    for (i=0; (i<length/sizeof(long)) && (flag == 1); i++) {
      if( *(ptr_test+i) != PATTERN + i )
        flag = 0;
    }
    if (flag == 1 )  
      printf("all %ld bytes are ok \n", length);
    else
      printf("\n - not all bytes are ok (%d -> %ld) \n\n", i, *(ptr_test+i) );
  }
}
