/* physmem_lib.c
 *
 * Library routines for ALICE RORC programs 
 *
 * last updated: 20/08/2011
 * written by: Ervin Denes
 *    
 * history: 
 *
 * v.5.1   20-Aug-2011: new rutine physmemBlockCross4B
 * v.4.13  03-Apr-2009: new physmemOpen()
 * v.4.11  14-Jan-2009: original version
 *   
 ****************************************************/
#define LIB_SOURCE_FILE "physmem_lib.c"
#define LIB_DESCRIPTION "ALICE physmem library"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <linux/types.h>
//#include <sys/io.h>
#include <sys/mman.h>
//#include <asm/page.h>
#include <sys/ioctl.h>

#include "physmem.h"
#include "physmem_version.h"
#include "physmem_lib.h"
char libraryIdent[]="@(#)""" LIB_SOURCE_FILE """: """ LIB_DESCRIPTION \
     """ """PHYSMEM_VERSION """ compiled """ __DATE__ """ """ __TIME__;

//#define arch64 (sizeof(void *) > 4)
#define memOffset(a, b) (unsigned long)((void *)a - (void *)b)

/*
 * Global variables ------------------------------------------------
 */


/*****************************************************************************/
int physmemOpen (physmemHandler_t *device, int minor, int full_size, 
                                           unsigned long requested_size)
{

  int physmem_fd;                   // file desciptor of physmem
  void *addr_user_physmem;          // pointer in user space to physmem
  unsigned long phys_size;
  unsigned long phys_addr_physmem;
  unsigned long phys_size_physmem;
  unsigned long phys_fullsize_physmem;
  unsigned long sum_gaps;
  int phys_num_zones;
  struct memZoneStruct *phys_zones;
  offZone_t *phys_offs;
  offZone_t *user_offs;
  int j;
  char physmem_name[16];

  device->zones = NULL;
  device->addr_user_physmem = NULL;

  if ((minor < 0) || (minor >= PHYSMEM_MAXDEVICE))
  {
    printf("Physmem minor %d is not valid\n", minor);
    return (EXIT_FAILURE);
  }

  sprintf(physmem_name, "%s%d", PHYSMEM_DEV, minor);
  physmem_fd = open(physmem_name,O_RDWR);
  if (physmem_fd  < 0)
  {
    perror("open");
    printf("Cannot open device %s \n", physmem_name);
    return (EXIT_FAILURE);
  }

  /* get physical address of physmem */
  if (ioctl (physmem_fd, PHYSMEM_GETADDR, &phys_addr_physmem) < 0)
  {
    printf("Cannot ioctl device %s to get address\n", physmem_name);
    close(physmem_fd);
    return (EXIT_FAILURE);
  }

  /* get size of the physmem */
  phys_size_physmem = 0;
  phys_fullsize_physmem = 0;
  if (full_size)
  {
    if (ioctl (physmem_fd, PHYSMEM_GETFULLSIZE, &phys_fullsize_physmem) < 0)
    {
      printf("Cannot ioctl device %s to get full size \n", physmem_name);
      close(physmem_fd);
      return (EXIT_FAILURE);
    }
    phys_size = phys_fullsize_physmem;
  }
  else
  {
    if (ioctl (physmem_fd, PHYSMEM_GETSIZE, &phys_size_physmem) < 0)
    {
      printf("Cannot ioctl device %s to get size \n", physmem_name);
      close(physmem_fd);
      return (EXIT_FAILURE);
    }
    phys_size = phys_size_physmem;
  }

  if ((requested_size == 0) || (requested_size > phys_size))
    requested_size = phys_size;

  /* get number of memory zones */
  phys_num_zones = 1;

  if (ioctl (physmem_fd, PHYSMEM_GETNUMMEMZONES, &phys_num_zones) < 0) {
    printf ("Cannot ioctl device %s to get hole address \n", physmem_name);
    close (physmem_fd);
    return (EXIT_FAILURE);
  }
  phys_zones = (struct memZoneStruct *) malloc (phys_num_zones * 
                                         sizeof(struct memZoneStruct));

  /* get memory zones */

  phys_zones[0].start = phys_addr_physmem;
  phys_zones[0].size = phys_size_physmem;
  phys_zones[0].end = phys_addr_physmem + phys_size_physmem;

  if (ioctl (physmem_fd, PHYSMEM_GETMEMZONES, phys_zones) < 0) {
    printf ("Cannot ioctl device %s to get memory zones \n", physmem_name);
    close (physmem_fd);
    return (EXIT_FAILURE);
  }

  /* calculate offsets */
  phys_offs = (offZone_t *) malloc (phys_num_zones * sizeof (offZone_t));
  user_offs =  (offZone_t *) malloc (phys_num_zones * sizeof (offZone_t));

  phys_offs[0].start = 0;
  phys_offs[0].end = phys_zones[0].size;
  user_offs[0].start = 0;
  user_offs[0].end = phys_zones[0].size;
  for (j = 1; j <phys_num_zones; j++)
  {
    phys_offs[j].start = phys_offs[j-1].end + phys_zones[j].start - phys_zones[j-1].end;
    phys_offs[j].end = phys_offs[j].start + phys_zones[j].size;
    if (full_size)
    {
      user_offs[j].start = phys_offs[j].start;
      user_offs[j].end = phys_offs[j].end;
    }
    else
    {
      user_offs[j].start = user_offs[j-1].end;
      user_offs[j].end = user_offs[j].start + phys_zones[j].size;
    }
  }

  /* map the whole physmem to user space */
   if ((addr_user_physmem = mmap(0, phys_size, PROT_READ|PROT_WRITE,
                                 MAP_SHARED, physmem_fd,0)) == MAP_FAILED)
  {
    perror("mmap failed");
    close(physmem_fd);
    return (EXIT_FAILURE);
  }

  /* calculate size of gaps */
  sum_gaps = 0;
  for (j = 1; j < phys_num_zones; j++)
  {
    sum_gaps += phys_zones[j].start - phys_zones[j-1].end;
  } 

  if (full_size)
  {
    /* calculate  phys_size_physmem */
    for (j = 0; j < phys_num_zones; j++)
      phys_size_physmem += phys_zones[j].size;

  }
  else
    phys_fullsize_physmem = phys_size_physmem + sum_gaps;

  strcpy(device->physmem_dev, physmem_name);
  device->fd = physmem_fd;
  device->minor = minor;
  device->full_size = full_size;
  device->num_zones = phys_num_zones;
  device->phys_addr_physmem = phys_addr_physmem;
  device->addr_user_physmem = addr_user_physmem;
  device->phys_size_physmem = phys_size_physmem;
  device->full_size_physmem = phys_fullsize_physmem;
  device->mapped_size_physmem = phys_size;
  device->sum_gaps = sum_gaps;
  device->zones = phys_zones;
  device->phys_offs = phys_offs;
  device->user_offs = user_offs;

  return 0;
}

/*****************************************************************************/
int physmemClose (physmemHandler_t device)
{
  unsigned long phys_size;

  /* free memory zones */
  if (device.zones != NULL)
  {
    free (device.zones); 
    free (device.phys_offs);
    free (device.user_offs);
  } 

  /* remove the mapping to physmem */
  if (device.addr_user_physmem != NULL)
  {
//    phys_size = device.full_size ? device.full_size_physmem : 
//                                   device.phys_size_physmem ;
    phys_size = device.mapped_size_physmem;                                      
    if (munmap((void *)device.addr_user_physmem, phys_size) < 0)
    {
      perror("physmem munmap failed");
      close(device.fd);
      return (EXIT_FAILURE);
    }
  }

  /* close the device for physmem */
  if (close (device.fd) == -1) 
  {
    printf("Cannot close physmem device\n");
    return (EXIT_FAILURE);
  }

  return 0;
}

/*****************************************************************************/
void physmemPrintZones(physmemHandler_t device)

{
  int j;
  int num_zones;
  struct memZoneStruct *zones;
  offZone_t *phys_offs;
  offZone_t *user_offs;

  num_zones = device.num_zones;
  zones = device.zones;
  phys_offs = device.phys_offs;
  user_offs = device.user_offs;

  printf("\n");
  printf (" + Number of memory zones : %d\n", num_zones);
  for (j = 0; j < num_zones; j++)
  {
    printf (" + Memory zone %d : %#16lx -> %#16lx, size: 0x%lx = %ld bytes (%.2f MB)\n",
         j+1, zones[j].start, zones[j].end, zones[j].size,
              zones[j].size, (double)zones[j].size/1024/1024);
    printf(" offsets physical: %#16lx -> %#16lx\n", phys_offs[j].start, phys_offs[j].end);
    printf("             user: %#16lx -> %#16lx\n", user_offs[j].start, user_offs[j].end);
  }
  printf (" + Total size of memory gaps: %ld bytes (%.2f MB)\n",
              device.sum_gaps, (double)device.sum_gaps/1024/1024);
  printf (" + Mapping method: %s\n", device.full_size ? "full size" :
                                                        "usable size");

}

/*****************************************************************************/
int physmemBlockIsNotContinuous(volatile unsigned int *block_start, 
                                volatile unsigned int *block_end,
                                physmemHandler_t      device,
                                int                   *start_status,
                                int                   *end_status)

/* The routine investigates if the block start and end is in the same
   memory zone or not. It returns 0 if they are in the same zone, -1
   otherwise.
   Other return values:
       start_status, if >= 0: number of memory zone, or 
                     if <  0: number of gap, where the block starts
       end_status: the same for the block end.

   Memory zone and gap numbers: 
   
    -1     0       -2      1        -3        2       -4 
        ----------     ---------         ----------     
   
*/
{
  int j;
  int i1, i2;
  unsigned long start_offset, end_offset;

  start_offset = memOffset(block_start, device.addr_user_physmem);
  end_offset = memOffset(block_end, device.addr_user_physmem);

  if (start_offset < device.user_offs[0].start)
  {
    i1 = -1;
  }
  if (end_offset < device.user_offs[0].start)
  {
    i2 = -1;
  }

  for (j = 0; j < device.num_zones; j++)
  {
    if ((start_offset >= device.user_offs[j].start) && 
        (start_offset < device.user_offs[j].end))
    {
      i1 = j;
    }
    if ((end_offset >= device.user_offs[j].start) &&
        (end_offset < device.user_offs[j].end))
    {
      i2 = j;
    }
  }

  for (j = 0; j < device.num_zones-1; j++)
  {
    if ((start_offset >= device.user_offs[j].end) && 
        (start_offset < device.user_offs[j+1].start))
    {
      i1 = -(j+2);
    }
    if ((end_offset >= device.user_offs[j].end) &&
        (end_offset < device.user_offs[j+1].start))
    {
      i2 = -(j+2);
    }
  }

  if (start_offset >= device.user_offs[device.num_zones-1].end)
  {
    i1 = -(device.num_zones + 1);
  }
  if (end_offset >= device.user_offs[device.num_zones-1].end)
  {
    i2 = -(device.num_zones + 1);
  }

  *start_status = i1;
  *end_status = i2;

  if ((i1 < 0) || (i2 < 0))  // block start or end is outside of zones
    return (-1);

  if (i1 != i2)              // block start and end are in different zones
     return (-1);

  return (0);
}

/*****************************************************************************/
// does block cross 4GB physicial address?
int physmemBlockCross4GB(unsigned long phys_address, long size)
{

//  if (!(sizeof(void *) > 4))
//    return (0);

#if __WORDSIZE > 32
  unsigned long high_start, high_end;
  high_start = phys_address >> 32;
  high_end = (phys_address + size - 1) >> 32;
  if (high_end != high_start) 
    return (1); 
  return (0);
#else
  return (0);
#endif
}

/*****************************************************************************/
// give physical offset from user offset 
unsigned long physmemPhysOffset(unsigned long user_offset,
                                physmemHandler_t device)
{
  int j;
  unsigned long phys_offset;

  if (device.full_size)
  {
    phys_offset = user_offset;
    return (phys_offset);
  }

  phys_offset = -1;
  for (j = 0; j < device.num_zones; j++)
  {
    if ((user_offset >= device.user_offs[j].start) && 
        (user_offset < device.user_offs[j].end))
    {
      phys_offset = device.phys_offs[j].start + 
                    (user_offset - device.user_offs[j].start);
      break;
    }
  }

  return (phys_offset);
}

/*****************************************************************************/
// give user offset from physical address 
unsigned long physmemUserOffset(unsigned long    phys_address,
                                physmemHandler_t device)
{
  int j;
  unsigned long user_offset;
  unsigned long phys_offset;

  if (phys_address < device.phys_addr_physmem)
    return (-1);    // phys address is under physmem memory

  phys_offset = phys_address - device.phys_addr_physmem;

  if (phys_offset > device.phys_offs[device.num_zones - 1].end)
    return (-1);   // phys address is over physmem memory

  if (device.full_size)
  {
    user_offset = phys_offset;
    return (user_offset);
  }
 
  user_offset = -1;
  for (j = 0; j < device.num_zones; j++)
  {
    if ((phys_offset >= device.phys_offs[j].start) && 
        (phys_offset < device.phys_offs[j].end))
    {
      user_offset = device.user_offs[j].start + 
                    (phys_offset - device.phys_offs[j].start);
      break;
    }
  }

  if (user_offset == -1)
  {
    for (j = 0; j < device.num_zones - 1; j++)
    {
      if ((phys_offset >= device.phys_offs[j].end) &&
          (phys_offset < device.phys_offs[j+1].start))
      {
        user_offset = -1;
        break;
      }
    }
  }

  return (user_offset);
}

